import torch
import torch.nn as nn, numpy as np
from torch.nn import functional as F
from torch_geometric.nn import MeanAggregation, GATv2Conv


class PlanGraphEmbedder(nn.Module):
    def __init__(self, device, h_dim, dropout=0.0, heads=4):
        super().__init__()
        self.device = device
        self.proj_x = nn.Sequential(
            nn.Linear(27, h_dim),
            nn.GELU(),
            nn.Dropout(dropout)
        )
        self.proj_edge_attr = nn.Sequential(
            nn.Linear(12, h_dim),
            nn.GELU(),
            nn.Dropout(dropout)
        )
        self.conv1 = GATv2Conv(h_dim, h_dim, heads=heads, edge_dim=h_dim)
        self.conv2 = GATv2Conv(h_dim*heads, h_dim, heads=1, edge_dim=h_dim)
        self.act = nn.GELU()
        self.dropout = nn.Dropout(dropout)
        self.dec = nn.Linear(h_dim*3, 1)

    def encode(self, data):
        x, edge_index, edge_attr = data.features.to(self.device), data.edge_index.to(self.device), data.tool.to(self.device)
        x = self.proj_x(x)
        edge_attr = self.proj_edge_attr(edge_attr)
        x = self.conv1(x, edge_index, edge_attr)
        x = self.act(x)
        x = self.dropout(x)
        x = self.conv2(x, edge_index, edge_attr)
        return x, edge_attr
    
    def decode(self, z, context, edge_label_index):
        u = z[edge_label_index[0]] 
        v = z[edge_label_index[1]] 
        return self.dec(torch.cat((u, v, context), -1))

    # def decode(self, z, edge_index, edge_attr, edge_label_index):
    #     z = self.conv3(z, edge_index.to(self.device), edge_attr)
    #     return (z[edge_label_index[0]] * z[edge_label_index[1]]).sum(dim=-1)

class Model(nn.Module):
    def __init__(self, seq_model_type, device):
        super(Model, self).__init__()
        
        self.device = device
        plan_emb_out = 128
        self.plan_embedder0 = PlanGraphEmbedder(device, plan_emb_out)
        self.plan_embedder1 = PlanGraphEmbedder(device, plan_emb_out)
        self.plan_embedder2 = PlanGraphEmbedder(device, plan_emb_out)
        self.plan_pool = MeanAggregation()
        dlist_hidden = 1024
        frame_emb = 512
        drnn_in = 5*1024 + 2 + frame_emb + 1024
        self.dialogue_listener_pre_ln = nn.LayerNorm(drnn_in)
        if seq_model_type==0:
            self.dialogue_listener_rnn = nn.GRU(drnn_in, dlist_hidden)
            self.dialogue_listener = lambda x: \
                self.dialogue_listener_rnn(x.reshape(-1, 1, drnn_in))[0]
        elif seq_model_type==1:
            self.dialogue_listener_rnn = nn.LSTM(drnn_in, dlist_hidden)
            self.dialogue_listener = lambda x: \
                self.dialogue_listener_rnn(x.reshape(-1, 1, drnn_in))[0]
        elif seq_model_type==2:
            mask_fun = lambda x: torch.triu(torch.ones(x.shape[0], x.shape[0]), diagonal=1).bool().to(device)
            sincos_fun = lambda x:torch.transpose(torch.stack([
                torch.sin(2*np.pi*torch.tensor(list(range(x)))/x),
                torch.cos(2*np.pi*torch.tensor(list(range(x)))/x)
                ]),0,1).reshape(-1,1,2)
            self.dialogue_listener_lin1 = nn.Linear(drnn_in, dlist_hidden-2)
            self.dialogue_listener_attn = nn.MultiheadAttention(dlist_hidden, 8)
            self.dialogue_listener_wrap = lambda x: self.dialogue_listener_attn(x, x, x, attn_mask=mask_fun(x))
            self.dialogue_listener = lambda x: self.dialogue_listener_wrap(torch.cat([
                sincos_fun(x.shape[0]).float().to(self.device),
                self.dialogue_listener_lin1(x).reshape(-1, 1, dlist_hidden-2)
            ], axis=-1))[0]
        else:
            print('Sequence model type must be in (0: GRU, 1: LSTM, 2: Transformer), but got ', seq_model_type)
            exit()
        conv_block = lambda i, o, k, p, s: nn.Sequential(
            nn.Conv2d(i, o, k, padding=p, stride=s),
            nn.BatchNorm2d(o),
            nn.Dropout(0.5),
            nn.MaxPool2d(2),
            nn.BatchNorm2d(o),
            nn.Dropout(0.5),
            nn.ReLU()
        )
        self.conv = nn.Sequential(
            conv_block(3, 8, 3, 1, 1),
            conv_block(8, 32, 5, 2, 2),
            conv_block(32, frame_emb//4, 5, 2, 2),
            nn.Conv2d(frame_emb//4, frame_emb, 3),
            nn.ReLU(),
        )
        self.proj_y = nn.Sequential(
            nn.Linear(1024, 512),
            nn.GELU(),
            nn.Dropout(0.5),
            nn.Linear(512, plan_emb_out)
        )
    
    def forward(self, game, experiment, global_plan=False, player_plan=False, incremental=False, return_feats=False):
        
        _, d, l, q, f, _, intermediate, _ = zip(*list(game))
        intermediate = np.array(intermediate)
        f = np.array(f, dtype=np.uint8)
        d = np.stack([np.concatenate(([int(x[0][1]==2), int(x[0][1]==1)], x[0][-1])) if not x is None else np.zeros(1026) for x in d])
        try:
            sel1 = int([x[0][2] for x in q if not x is None][0] == 1)
            sel2 = 1 - sel1
        except Exception as e:
            sel1 = 0
            sel2 = 0
        
        if player_plan:
            if sel1: 
                z, _ = self.plan_embedder1.encode(game.player1_plan)
            elif sel2:
                z, _ = self.plan_embedder2.encode(game.player2_plan)
            else: 
                z, _ = self.plan_embedder0.encode(game.global_plan)
        else:
            raise ValueError('There should never be a global plan!')

        u = torch.cat((
            torch.tensor(d).float().to(self.device),
            self.conv(torch.tensor(f).permute(0, 3, 1, 2).float().to(self.device) / 255.0).reshape(-1, 512),
            torch.tensor(intermediate).float().to(self.device)
        ), axis=-1)
        u = u.float().to(self.device)
        u = self.dialogue_listener_pre_ln(u)
        y = self.dialogue_listener(u)
        y = y.reshape(-1, y.shape[-1])
        if return_feats:
            _y = y.clone().detach().cpu().numpy()
        y = self.proj_y(y)

        if experiment == 2:
            pred, label = self.decode_own_missing_knowledge(z, y, game, sel1, sel2, incremental)
        elif experiment == 3:
            pred, label = self.decode_partner_missing_knowledge(z, y, game, sel1, sel2, incremental) 
        else:
            raise ValueError('Wrong experiment id! Valid values are 2 and 3.')
        
        if return_feats:
            return pred, label, [sel1, sel2], _y 
        
        return pred, label, [sel1, sel2]
        
    def decode_own_missing_knowledge(self, z, y, game, sel1, sel2, incremental):
        if incremental:
            if sel1:
                pred = torch.stack(
                    [self.plan_embedder1.decode(z, y[0].repeat(game.player1_edge_label_index.shape[1], 1), game.player1_edge_label_index).view(-1)] 
                    + 
                    [self.plan_embedder1.decode(z, f.repeat(game.player1_edge_label_index.shape[1], 1), game.player1_edge_label_index).view(-1) for f in y[len(y)%10-1::10]]
                )
                label = game.player1_edge_label_own_missing_knowledge.to(self.device)
            elif sel2: 
                pred = torch.stack(
                    [self.plan_embedder2.decode(z, y[0].repeat(game.player2_edge_label_index.shape[1], 1), game.player2_edge_label_index).view(-1)] 
                    + 
                    [self.plan_embedder2.decode(z, f.repeat(game.player2_edge_label_index.shape[1], 1), game.player2_edge_label_index).view(-1) for f in y[len(y)%10-1::10]]
                )
                label = game.player2_edge_label_own_missing_knowledge.to(self.device)
            else: 
                pred = torch.stack(
                    # NOTE: here I use game.global_plan.edge_index instead of edge_label_index => no negative sampling
                    [self.plan_embedder0.decode(z, y[0].repeat(game.global_plan.edge_index.shape[1], 1), game.global_plan.edge_index).view(-1)] 
                    + 
                    [self.plan_embedder0.decode(z, f.repeat(game.global_plan.edge_index.shape[1], 1), game.global_plan.edge_index).view(-1) for f in y[len(y)%10-1::10]]
                )
                label = torch.zeros(game.global_plan.edge_index.shape[1])
        else:
            if sel1: 
                pred = self.plan_embedder1.decode(z, y.mean(0, keepdim=True).repeat(game.player1_edge_label_index.shape[1], 1), game.player1_edge_label_index).view(-1)
                label = game.player1_edge_label_own_missing_knowledge.to(self.device)
            elif sel2:
                pred = self.plan_embedder2.decode(z, y.mean(0, keepdim=True).repeat(game.player2_edge_label_index.shape[1], 1), game.player2_edge_label_index).view(-1)
                label = game.player2_edge_label_own_missing_knowledge.to(self.device)
            else:
                # NOTE: here I use game.global_plan.edge_index instead of edge_label_index => no negative sampling
                pred = self.plan_embedder0.decode(z, y.mean(0, keepdim=True).repeat(game.global_plan.edge_index.shape[1], 1), game.global_plan.edge_index).view(-1)
                label = torch.zeros(game.global_plan.edge_index.shape[1])
        
        return (pred, label)
    
    def decode_partner_missing_knowledge(self, z, y, game, sel1, sel2, incremental):
        if incremental:
            if sel1:
                pred = torch.stack(
                    [self.plan_embedder1.decode(z, y[0].repeat(game.player1_plan.edge_index.shape[1], 1), game.player1_plan.edge_index).view(-1)] 
                    + 
                    [self.plan_embedder1.decode(z, f.repeat(game.player1_plan.edge_index.shape[1], 1), game.player1_plan.edge_index).view(-1) for f in y[len(y)%10-1::10]]
                )
                label = game.player1_edge_label_other_missing_knowledge.to(self.device)
            elif sel2: 
                pred = torch.stack(
                    [self.plan_embedder2.decode(z, y[0].repeat(game.player2_plan.edge_index.shape[1], 1), game.player2_plan.edge_index).view(-1)] 
                    + 
                    [self.plan_embedder2.decode(z, f.repeat(game.player2_plan.edge_index.shape[1], 1), game.player2_plan.edge_index).view(-1) for f in y[len(y)%10-1::10]]
                )
                label = game.player2_edge_label_other_missing_knowledge.to(self.device)
            else: 
                pred = torch.stack(
                    # NOTE: here I use game.global_plan.edge_index instead of edge_label_index => no negative sampling
                    [self.plan_embedder0.decode(z, y[0].repeat(game.global_plan.edge_index.shape[1], 1), game.global_plan.edge_index).view(-1)] 
                    + 
                    [self.plan_embedder0.decode(z, f.repeat(game.global_plan.edge_index.shape[1], 1), game.global_plan.edge_index).view(-1) for f in y[len(y)%10-1::10]]
                )
                label = torch.zeros(game.global_plan.edge_index.shape[1])
        else:
            if sel1: 
                pred = self.plan_embedder1.decode(z, y.mean(0, keepdim=True).repeat(game.player1_plan.edge_index.shape[1], 1), game.player1_plan.edge_index).view(-1)
                label = game.player1_edge_label_other_missing_knowledge.to(self.device)
            elif sel2:
                pred = self.plan_embedder2.decode(z, y.mean(0, keepdim=True).repeat(game.player2_plan.edge_index.shape[1], 1), game.player2_plan.edge_index).view(-1)
                label = game.player2_edge_label_other_missing_knowledge.to(self.device)
            else:
                # NOTE: here I use game.global_plan.edge_index instead of edge_label_index => no negative sampling
                pred = self.plan_embedder0.decode(z, y.mean(0, keepdim=True).repeat(game.global_plan.edge_index.shape[1], 1), game.global_plan.edge_index).view(-1)
                label = torch.zeros(game.global_plan.edge_index.shape[1])
        
        return (pred, label)
    